#!/bin/bash
# OCR.sh
# multithreaded OCR with tesseract 3.02, tested in Ubuntu 12.04
# pokes text as layer into PDF, works in memory (/run/shm/)
# see convert, tesseract and hocr2pdf man pages for available options
# Michael Luthardt <edv@dr-luthardt.de> 2013, 2014

# check for valid PDF as first argument
if (! [ $# -ge 1 ] || ! (gs -q -o nul -sDEVICE=nullpage -dFirstPage=1 -dLastPage=1 "$1" &>/dev/null)); then
	cat << eot
	First argument is not a valid PDF or doesn't exist.
	Usage: [path/to/]pdffile [lan] [-y]
		See tesseract --list-langs for installed languages.
eot
	exit 1
fi

# check for parameters
# language as 2nd argument, if not, set default (deu)
# it's your own task to provide lan.traineddata
# with option -y OCR is forced anyway

FORCE=0 # as a precaution
if [ ${#} -eq 1 ]; then
	LANG=deu
else
	[ ${#2} -eq 3 ] && LANG=$2 || LANG=deu
	[ "x$2" == "x-y" ] && FORCE=1
	[ "x$3" == "x-y" ] && FORCE=1
fi

if [ $FORCE -eq 0 ]; then
#check if PDF already contains a text layer
	unset -v ANS	# as a precaution
	if [ $(pdftotext "$1" - | grep -cE '[[:alpha:]]' 2>/dev/null) -ne 0 ]; then
		echo -e "\nThis PDF contains a text layer."
		read -r -p "Proceed anyway? Old text will be removed. [N|y] " ANS
		[ "x$ANS" = "xy" ] || { echo; exit 2; }
	fi
fi

# get names
if [ "$(dirname "$1")" = "." ]; then
	DIR=`pwd`
else
	DIR=`dirname "$1"`
fi
FILE=`basename "$1"`

# find number of cpus, calculate maxjobs
NCPUS=$(lscpu | sed -n '/^CPU(s)/s/CPU(s): *//p')
MAXJOBS=$((($NCPUS+1)/2))

# to work in memory:
# ------------------
# copy the PDF into it
cp "$1" /run/shm/
# change the working directory
cd /run/shm/
# redirect magick convert
export MAGICK_TEMPORARY_PATH=/run/shm
# and its gs backend temporary path
export TEMPDIR=/run/shm
# -------------------

# burst input PDF into pages
rm -f pg_* # as a precaution
pdftk "$FILE" burst
tput bold; echo -e "\n${FILE}:"; tput sgr0
echo "`sed -n '/NumberOfPages/s/NumberOfPages: //p' doc_data.txt` pages to process ..."
# we don't need the input PDF and it's info doc any longer in memory
rm -f "$FILE" doc_data.txt

# bundle all actions into one function
export ERR=0
ocr() 
{
    convert -density 300x300 $PAGE ${PAGE%.*}.png &>/dev/null
    let ERR=$ERR+$?
    tesseract ${PAGE%.*}.png ${PAGE%.*} -l $LANG hocr &>/dev/null
    let ERR=$ERR+$?
    convert ${PAGE%.*}.png ${PAGE%.*}.bmp &>/dev/null
    let ERR=$ERR+$?
    hocr2pdf -i ${PAGE%.*}.bmp -o ${PAGE%.*}.pdf < ${PAGE%.*}.html &>/dev/null
    let ERR=$ERR+$?
    [ $ERR -eq 0 ] && { tput civis; echo -en "        ... $PAGE done\r"; tput cnorm; } \
    || { tput civis; echo -en "\n        ... $PAGE something went wrong\n"; tput cnorm; }
    rm -f ${PAGE%.*}.png ${PAGE%.*}.bmp ${PAGE%.*}.html 
}

# do orc() in background
for PAGE in pg_*.pdf; do
	ocr $PAGE &
	# but limit the number of simultaneous jobs
 	[ `jobs -p | wc -l` -ge $MAXJOBS ] && wait
 	[ $ERR -ne 0 ] && break
done

# wait for last bg job to finish – important!
wait

# concatenate pdf's to input_ocr.pdf
if [ $ERR -eq 0 ]; then
	OUTFILE="${DIR}/${FILE}"
	pdftk pg_*.pdf cat output "${OUTFILE%.*}_ocr.pdf" &>/dev/null
	[ $? -eq 0 ] && echo -e "\n\t... ${OUTFILE%.*}_ocr.pdf created\n" \
	|| echo -e "\n\t... Errors encountered.  No output created.\n"
else
	echo -e "\n\t... Errors encountered.  No output created.\n"
fi

# clear memory
rm -f pg_*

exit
